#############################################################################
##
#W    simplicialhomology.gi   share package 'homology'      Dumas, Heckenbach
##
##==========================================================
## Homology Groups in GAP 4
## (c) July 1999
##==========================================================

#==========================================================
# Information level
#==========================================================

DeclareInfoClass ("InfoHomology");

InstallGlobalFunction (HomologyInfo, function (level)
  SetInfoLevel (InfoHomology, level);
end);

HomologyInfo (10); # default info level

#==================================================
# Error Tolerance 
#==================================================
InstallGlobalFunction (SetUncertaintyTolerance, function (error_tol)
  local et;
  et := 0;
  if error_tol in Rationals then
	if error_tol > 1 then
	        et := 1/error_tol;
               	Info (InfoWarning, 1, "Error tolerance setted to the inverse.");
	else if error_tol >= 0 then
		et := error_tol;
             else
		Info (InfoWarning, 1, "Invalid error tolerance! Using the default.");
	     fi;
	fi;
  else
	Info (InfoWarning, 1, "Invalid error tolerance! Using the default.");
  fi;
  if error_tol = 1 then
	et := 0; # Internal default is 0 (even if external default is 1)
  fi;
  PushOptions (rec (UncertaintyTolerance := et));
end);


#==================================================
# Available algorithms
#==================================================

# old values 
# InstallValue (HomologyAlgorithms, ["Integer", "Integer-GMP", "Valence-elim", "Valence-bb"]);
InstallValue (HomologyAlgorithms, ["EliminateAlgorithm", "EliminateGMPAlgorithm", "ValenceElimAlgorithm", "ValenceBBAlgorithm"]);
InstallValue (EliminateAlgorithm,    HomologyAlgorithms [1]);
InstallValue (EliminateGMPAlgorithm, HomologyAlgorithms [2]);
InstallValue (ValenceElimAlgorithm,  HomologyAlgorithms [3]);
InstallValue (ValenceBBAlgorithm,    HomologyAlgorithms [4]);

InstallGlobalFunction (SetHomologyAlgorithm, function (Algorithm)
  if Algorithm in HomologyAlgorithms then
    PushOptions (rec (HomologyAlgorithm := Algorithm));
  else
    Info (InfoWarning, 1, "Invalid homology algorithm set! Using the default.");
  fi;
end);

#==================================================
# Available output formats
#==================================================

InstallValue (HomologyGroupFormats, ["CyclicPSubgroups", "MaximalCyclicSubgroups", "GapFormat"]);
InstallValue (CyclicPSubgroups,       HomologyGroupFormats [1]);
InstallValue (MaximalCyclicSubgroups, HomologyGroupFormats [2]);
InstallValue (GapFormat,              HomologyGroupFormats [3]);

InstallGlobalFunction (SetHomologyGroupFormat, function (Format)
  if Format in HomologyGroupFormats then
    PushOptions (rec (HomologyGroupFormat := Format));
  else
    Info (InfoWarning, 1, "Invalid homology group format set! Using the default.");
  fi;
end);

#==================================================
# A Simplicial complex is a dense list of dense lists
#==================================================

InstallGlobalFunction (IsSimplicialComplex, function (complex)
  return IsDenseList (complex) and ForAll (complex, IsDenseList);
end);

#==================================================
# Simplicial Homology
# Internal function: Parsing arguments and choosing between the executables
#==================================================

ComputeSimplicialHomology := function (args, procname, mode, coeff)
  local maxargs, nargs, complex, range, min, max, AlgStr, Algorithm, FormatStr, Format,
        tmpdir, filename, streamin, streamout, arguments, fname, s, Result, Status, Usage,
        MaximalCyclicSubgroupsToCyclicPSubgroups, MaximalCyclicSubgroupsToGapFormat, ErrorTolerance, ETStr;

  MaximalCyclicSubgroupsToCyclicPSubgroups := function (x)
    local l;
    l := Length (x);
    if l = 1 then
      return x;
    else
      return Concatenation ([x [1]], Concatenation (List (Unique (FactorsInt (x [l][1])),
        function (p)
          local a, i, e;
          a := [];
          for i in [First ([2 .. l], j -> RemInt (x [j][1], p) = 0) .. l] do
            e := Number (FactorsInt (x [i][1]), k -> k = p);
            if IsBound (a [e]) then
              a [e] := a [e] + x [i][2];
            else
              a [e] := x [i][2];
            fi;
          od;
          return List (Filtered ([1 .. Length (a)], i -> IsBound (a [i])),
                       k -> [p ^ k, a [k]]);
        end)));
    fi;
  end;

  MaximalCyclicSubgroupsToGapFormat := function (x)
    local l;
    l := Length (x);
    return [x [1], Concatenation (List (
      x {[2 .. l]}, t -> ListWithIdenticalEntries (t [2], t [1])))];
  end;

  Usage := function ()
    Print ("usage: ", procname, " (<complex>");
    if mode = 2 then Print (", <torsion>"); fi;
    if mode <> 3 and mode <> 4 then
      Print (" [,<dimensions>]);\n");
    else
      Print (" [,<dimensions>] [,<filename>]);\n");
    fi;
    Print ("where:\n");
    Print ("complex    : a simplicial complex (dense list of dense lists)\n");
    if mode = 2 then
      Print ("torsion    : the coefficient of the torsion ranks wanted\n");
    fi;
    Print ("dimensions : specification of the homology groups to compute, this may be:\n");
    Print ("  - an integer, to compute one group,\n");
    Print ("  - a pair of integers, to compute a range of groups or,\n");
    Print ("  - the empty list (default), to compute all groups\n");
    if mode = 3 or mode = 4 then
      Print ("filename   : output file for the matrix (in sparse matrix format)\n");
    fi;
    if mode <> 3 and mode <> 4 then
      Print ("\n");
      Print ("The Option \"HomologyAlgorithm\" can select the algorithm to be used:\n");
      Print ("  - \"Integer\"      : Integer elimination with gcd (default)\n");
      Print ("  - \"Integer-GMP\"  : Integer elimination with gcd, using\n");
      Print ("                     multi-precision arithmetic (GMP library)\n");
      Print ("  - \"Valence-elim\" : Valence, via Laplacians and ranks mod m,\n");
      Print ("                     via elimination\n");
      Print ("  - \"Valence-bb\"   : Valence, via Laplacians and ranks mod m,\n");
      Print ("                     via blackbox methods\n");
      Print ("The Option \"UncertaintyTolerance\" can set the error tolerance for the Valence algorithms:\n");
      Print ("  - 0			: The algorithm chooses some tolerance (default)\n");
      Print ("  - 1			: The algorithm chooses some tolerance\n");
      Print ("  - a Rational r, 0<r<1	: Takes r as error tolerance\n");
      Print ("  - a Rational r, r>1	: Takes 1/r as error tolerance\n");
      Print ("  - an Integer i, i!=0	: Takes 1/i as error tolerance\n");
    fi;
    if mode = 0 then
      Print ("\n");
      Print ("The Option \"HomologyGroupFormat\" can select the format for the\n");
      Print ("representation of a homology group:\n");
      Print ("  - \"CyclicPSubgroups\"       : Direct product of maximal cyclic\n");
      Print ("                               p-subgroups (default)\n");
      Print ("  - \"MaximalCyclicSubgroups\" : Direct product of maximal cyclic subgroups\n");
      Print ("  - \"GapFormat\"              : Direct product of maximal cyclic subgroups,\n");
      Print ("                               listed such that the torsion part can be used\n");
      Print ("                               as input to the Gap-function `AbelianGroup'.\n");
    fi;
  end;

  nargs := Length (args);
  if mode = 4 then
    maxargs := 3;
  else
    maxargs := 2;
  fi;
  if (nargs < 1) or (nargs > maxargs) then
    Info (InfoWarning, 1, "Error : Invalid number of arguments\n");
    Usage ();
    return fail;
  fi;
  if mode = 2 and (not (coeff in Integers) or (coeff < 2)) then
    Info (InfoWarning, 1, "Torsion coefficient must be an integer >= 2.");
    return fail;
  fi;
  complex := args [1];
  if not IsSimplicialComplex (complex) then
    Info (InfoWarning, 1, "First argument must be a simplicial complex (dense list of dense lists).");
    return fail;
  fi;
  min := 0;
  max := -1;
  if nargs >= 2 then
    range := args[2];
    if range in Integers then
      min := range;
      max := range;
      if max < 0 then return [[]]; fi;
    elif IsList (range) and Length (range) = 2
      and IsBound (range [1]) and range [1] in Integers
      and IsBound (range [2]) and range [2] in Integers then
      min := range [1];
      max := range [2];
      if max < 0 then return [[]]; fi;
    elif IsList (range) and Length (range) = 0 then
      # default
    else
      Info (InfoWarning, 2, "invalid range of groups, taking the default.");
    fi;
  fi;

  filename:= "";
  if mode = 4 then
    if not (range in Integers) then
      Info (InfoWarning, 1, "In file only one map at a time is generated, generating the first one.");
      max := min;
    fi;
    if not IsString (args[3]) then
      Info (InfoWarning, 1, "Third argument must be a string, used as an output filename.");
      return fail;
    fi;
    filename := args[3];
  fi;

  Algorithm := 1;
  AlgStr := ValueOption ("HomologyAlgorithm");
  if AlgStr <> fail then
    Algorithm := First ([1 .. Length (HomologyAlgorithms)], i -> HomologyAlgorithms [i] = AlgStr);
    if Algorithm = fail then
      Algorithm := 1;
      Info (InfoWarning, 1, "Invalid homology algorithm set! Using the default.");
    fi;
  fi;

  ErrorTolerance := 0;
  ETStr := ValueOption("UncertaintyTolerance");
  if ETStr <> fail then
    ErrorTolerance := ETStr;
    if ErrorTolerance = fail then
	ErrorTolerance := 0;
    fi;
  fi;

  Format := 1;
  FormatStr := ValueOption ("HomologyGroupFormat");
  if FormatStr <> fail then
    Format := First ([1 .. Length (HomologyGroupFormats)], i -> HomologyGroupFormats [i] = FormatStr);
    if Format = fail then
      Format := 1;
      Info (InfoWarning, 1, "Invalid homology group format set! Using the default.");
    fi;
  fi;

  # Filter out trivialities
  complex := Filtered (complex, s -> not IsEmpty (s));
  if Length (complex) = 0 then
    return [];
  fi;

  arguments := [String (min), String (max), String (InfoLevel (InfoHomology)), String (InfoLevel (InfoTiming)), String (Algorithm), String (mode), String (coeff), String(NumeratorRat(ErrorTolerance)), String(DenominatorRat(ErrorTolerance)), filename];
  tmpdir := DirectoryTemporary ();
  fname := Filename (tmpdir, "ftmp");
  streamout := OutputTextFile (fname, false);
  WriteLine (streamout, String (Elements (List (complex, s -> Elements (s)))));
  CloseStream (streamout);
  streamin := InputTextFile (fname);
  s := "return ";
  streamout := OutputTextString (s, true);
  Status := Process (DirectoryCurrent (), Filename (DirectoriesPackagePrograms ("homology")[2], "homology_gap"), streamin, streamout, arguments);
  WriteLine (streamout, ";");
  CloseStream (streamout);
  CloseStream (streamin);
  if Status <> 0 then
    return fail;
  else
    if mode <> 4 then
      Result := ReadAsFunction (InputTextString (s)) ();
      if mode = 0 then
        if Format = 1 then
          Result := List (Result, MaximalCyclicSubgroupsToCyclicPSubgroups);
        elif Format = 3 then
          Result := List (Result, MaximalCyclicSubgroupsToGapFormat);
        fi;
      fi;
      return Result;
    else
      return true;
    fi;
  fi;
end;

#==================================================
# Computing information about simplicial complexes
#==================================================

InstallGlobalFunction (SimplicialHomology, function (arg)
  return ComputeSimplicialHomology (arg, "SimplicialHomology", 0, 0);
end);

InstallGlobalFunction (SimplicialBettiNumbers, function (arg)
  local Temp;
  Temp := ComputeSimplicialHomology (arg, "SimplicialBettiNumbers", 1, 0);
  if Temp = fail then
    return fail;
  else
    return List (Temp, g -> g[1]);
  fi;
end);

InstallGlobalFunction (SimplicialNTorsionRanks, function (arg)
  local Temp, p;
  if Length (arg) < 2 then
    p := 1;
    arg := [];
  else
    p := arg [2];
    arg := List (Concatenation ([1], [3 .. Length (arg)]), i -> arg [i]);
  fi;
  Temp := ComputeSimplicialHomology (arg, "SimplicialNTorsionRanks", 2, p);
  if Temp = fail then
    return fail;
  else
    return List (Temp, g ->
      Sum ([2 .. Length (g)],
        function (i)
          if g [i] [1] mod p = 0 then
            return g [i] [2];
          else
            return 0;
          fi;
        end));
  fi;
end);

InstallGlobalFunction (SimplicialBoundaryMaps, function (arg)
  if Length (arg) < 3 then
    return ComputeSimplicialHomology (arg, "SimplicialBoundaryMaps", 3, 0);
  else
    ComputeSimplicialHomology (arg, "SimplicialBoundaryMaps", 4, 0);
  fi;
end);

InstallGlobalFunction (SimplicialComplexDimension, function (complex)
  if IsSimplicialComplex (complex) then
    return Maximum (List (complex, s -> Length (s))) - 1;
  else
    return fail;
  fi;
end);
